import uuid
from unittest.mock import Mock, patch
from datetime import datetime, timezone, timedelta

import pytest
from stellar_sdk import Keypair
from rest_framework.request import Request

from polaris.tests.helpers import mock_check_auth_success
from polaris.models import Transaction, OffChainAsset, ExchangePair, Quote
from polaris.sep31.transactions import process_post_response


transaction_endpoint = "/sep31/transactions"


success_send_integration = Mock(
    info=Mock(
        return_value={
            "fields": {
                "transaction": {"bank_account": {"description": "bank account"}}
            },
        }
    ),
    process_post_request=Mock(return_value=None),
)


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch(
    "polaris.sep31.transactions.registered_sep31_receiver_integration",
    success_send_integration,
)
def test_successful_send(client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31])
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "amount": 100,
            "sender_id": "123",
            "receiver_id": "456",
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 201
    assert all(
        f in body
        for f in ["id", "stellar_account_id", "stellar_memo_type", "stellar_memo"]
    )
    assert body["stellar_memo_type"] == Transaction.MEMO_TYPES.hash
    assert body["stellar_account_id"] == asset.distribution_account
    kwargs = success_send_integration.info.call_args_list[-1][1]
    assert isinstance(kwargs.get("request"), Request)
    assert kwargs.get("asset").code == asset.code
    assert kwargs.get("asset").issuer == asset.issuer
    assert kwargs.get("lang") is None
    success_send_integration.process_post_request.assert_called_once()
    kwargs = success_send_integration.process_post_request.call_args[1]
    success_send_integration.process_post_request.reset_mock()
    transaction = kwargs.get("transaction")
    assert isinstance(transaction, Transaction)
    assert transaction.amount_in == 100
    assert transaction.asset.code == asset.code
    assert transaction.kind == Transaction.KIND.send
    assert transaction.protocol == Transaction.PROTOCOL.sep31
    assert transaction.receiving_anchor_account == asset.distribution_account


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch(
    "polaris.sep31.transactions.registered_sep31_receiver_integration",
    success_send_integration,
)
def test_successful_send_indicative_quote(client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31, "sep38"])
    offchain_asset = OffChainAsset.objects.create(scheme="test", identifier="test")
    ExchangePair.objects.create(
        sell_asset=asset.asset_identification_format,
        buy_asset=offchain_asset.asset_identification_format,
    )
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "destination_asset": offchain_asset.asset_identification_format,
            "amount": 100,
            "sender_id": "123",
            "receiver_id": "456",
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 201, response.content
    assert all(
        f in body
        for f in ["id", "stellar_account_id", "stellar_memo_type", "stellar_memo"]
    )
    assert body["stellar_memo_type"] == Transaction.MEMO_TYPES.hash
    assert body["stellar_account_id"] == asset.distribution_account
    kwargs = success_send_integration.info.call_args_list[-1][1]
    assert isinstance(kwargs.get("request"), Request)
    assert kwargs.get("asset").code == asset.code
    assert kwargs.get("asset").issuer == asset.issuer
    assert kwargs.get("lang") is None
    success_send_integration.process_post_request.assert_called_once()
    kwargs = success_send_integration.process_post_request.call_args[1]
    success_send_integration.process_post_request.reset_mock()
    transaction = kwargs.get("transaction")
    assert isinstance(transaction, Transaction)
    assert transaction.amount_in == 100
    assert transaction.asset.code == asset.code
    assert transaction.kind == Transaction.KIND.send
    assert transaction.protocol == Transaction.PROTOCOL.sep31
    assert transaction.receiving_anchor_account == asset.distribution_account
    assert isinstance(transaction.quote, Quote)
    assert transaction.quote.id
    assert transaction.quote.type == Quote.TYPE.indicative
    assert transaction.quote.sell_asset == asset.asset_identification_format
    assert transaction.quote.buy_asset == offchain_asset.asset_identification_format
    assert transaction.quote.sell_amount == transaction.amount_in


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch(
    "polaris.sep31.transactions.registered_sep31_receiver_integration",
    success_send_integration,
)
def test_successful_send_firm_quote(client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31, "sep38"])
    offchain_asset = OffChainAsset.objects.create(scheme="test", identifier="test")
    ExchangePair.objects.create(
        sell_asset=asset.asset_identification_format,
        buy_asset=offchain_asset.asset_identification_format,
    )
    quote = Quote.objects.create(
        id=str(uuid.uuid4()),
        stellar_account="test source address",
        type=Quote.TYPE.firm,
        sell_asset=asset.asset_identification_format,
        buy_asset=offchain_asset.asset_identification_format,
        sell_amount=100,
        buy_amount=100,
        price=1,
        expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
    )
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "destination_asset": offchain_asset.asset_identification_format,
            "quote_id": quote.id,
            "amount": 100,
            "sender_id": "123",
            "receiver_id": "456",
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 201, response.content
    assert all(
        f in body
        for f in ["id", "stellar_account_id", "stellar_memo_type", "stellar_memo"]
    )
    assert body["stellar_memo_type"] == Transaction.MEMO_TYPES.hash
    assert body["stellar_account_id"] == asset.distribution_account
    kwargs = success_send_integration.info.call_args_list[-1][1]
    assert isinstance(kwargs.get("request"), Request)
    assert kwargs.get("asset").code == asset.code
    assert kwargs.get("asset").issuer == asset.issuer
    assert kwargs.get("lang") is None
    success_send_integration.process_post_request.assert_called_once()
    kwargs = success_send_integration.process_post_request.call_args[1]
    success_send_integration.process_post_request.reset_mock()
    transaction = kwargs.get("transaction")
    assert isinstance(transaction, Transaction)
    assert transaction.amount_in == 100
    assert transaction.asset.code == asset.code
    assert transaction.kind == Transaction.KIND.send
    assert transaction.protocol == Transaction.PROTOCOL.sep31
    assert transaction.receiving_anchor_account == asset.distribution_account
    assert isinstance(transaction.quote, Quote)
    assert str(transaction.quote.id) == str(quote.id)


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration")
def test_send_indicative_quote_not_enabled(mock_sep31_ri, client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31])
    offchain_asset = OffChainAsset.objects.create(scheme="test", identifier="test")
    ExchangePair.objects.create(
        sell_asset=asset.asset_identification_format,
        buy_asset=offchain_asset.asset_identification_format,
    )
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "destination_asset": offchain_asset.asset_identification_format,
            "amount": 100,
            "sender_id": "123",
            "receiver_id": "456",
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 400, response.content
    assert body == {"error": "quotes are not supported"}
    mock_sep31_ri.process_post_request.assert_not_called()


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration")
def test_send_indicative_quote_no_exchange_pair(
    mock_sep31_ri, client, usd_asset_factory
):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31, "sep38"])
    offchain_asset = OffChainAsset.objects.create(scheme="test", identifier="test")
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "destination_asset": offchain_asset.asset_identification_format,
            "amount": 100,
            "sender_id": "123",
            "receiver_id": "456",
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 400, response.content
    assert body == {
        "error": "unsupported 'destination_asset' for 'asset_code' and 'asset_issuer'"
    }
    mock_sep31_ri.process_post_request.assert_not_called()


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration")
def test_send_indicative_quote_no_offchain_asset(
    mock_sep31_ri, client, usd_asset_factory
):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31, "sep38"])
    ExchangePair.objects.create(
        sell_asset=asset.asset_identification_format, buy_asset="test:test"
    )
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "destination_asset": "test:test",
            "amount": 100,
            "sender_id": "123",
            "receiver_id": "456",
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 400, response.content
    assert body == {"error": "invalid 'destination_asset'"}
    mock_sep31_ri.process_post_request.assert_not_called()


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration")
def test_send_firm_quote_not_supported(mock_sep31_ri, client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31])
    offchain_asset = OffChainAsset.objects.create(scheme="test", identifier="test")
    ExchangePair.objects.create(
        sell_asset=asset.asset_identification_format,
        buy_asset=offchain_asset.asset_identification_format,
    )
    quote = Quote.objects.create(
        id=str(uuid.uuid4()),
        stellar_account="test source address",
        type=Quote.TYPE.firm,
        sell_asset=asset.asset_identification_format,
        buy_asset=offchain_asset.asset_identification_format,
        sell_amount=100,
        buy_amount=100,
        price=1,
        expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
    )
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "destination_asset": offchain_asset.asset_identification_format,
            "quote_id": quote.id,
            "amount": 100,
            "sender_id": "123",
            "receiver_id": "456",
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 400, response.content
    assert body == {"error": "quotes are not supported"}
    mock_sep31_ri.process_post_request.assert_not_called()


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration")
def test_send_firm_quote_no_destination_asset(mock_sep31_ri, client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31, "sep38"])
    offchain_asset = OffChainAsset.objects.create(scheme="test", identifier="test")
    ExchangePair.objects.create(
        sell_asset=asset.asset_identification_format,
        buy_asset=offchain_asset.asset_identification_format,
    )
    quote = Quote.objects.create(
        id=str(uuid.uuid4()),
        stellar_account="test source address",
        type=Quote.TYPE.firm,
        sell_asset=asset.asset_identification_format,
        buy_asset=offchain_asset.asset_identification_format,
        sell_amount=100,
        buy_amount=100,
        price=1,
        expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
    )
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "quote_id": quote.id,
            "amount": 100,
            "sender_id": "123",
            "receiver_id": "456",
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 400, response.content
    assert body == {
        "error": "'destination_asset' must be provided if 'quote_id' is provided"
    }
    mock_sep31_ri.process_post_request.assert_not_called()


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration")
def test_send_firm_quote_unknown_destination_asset(
    mock_sep31_ri, client, usd_asset_factory
):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31, "sep38"])
    offchain_asset = OffChainAsset.objects.create(scheme="test", identifier="test")
    ExchangePair.objects.create(
        sell_asset=asset.asset_identification_format,
        buy_asset=offchain_asset.asset_identification_format,
    )
    quote = Quote.objects.create(
        id=str(uuid.uuid4()),
        stellar_account="test source address",
        type=Quote.TYPE.firm,
        sell_asset=asset.asset_identification_format,
        buy_asset=offchain_asset.asset_identification_format,
        sell_amount=100,
        buy_amount=100,
        price=1,
        expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
    )
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "quote_id": quote.id,
            "destination_asset": "not:test",
            "amount": 100,
            "sender_id": "123",
            "receiver_id": "456",
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 400, response.content
    assert body == {
        "error": "quote 'buy_asset' does not match 'destination_asset' parameter"
    }
    mock_sep31_ri.process_post_request.assert_not_called()


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration")
def test_send_firm_quote_filters_for_firm_quotes(
    mock_sep31_ri, client, usd_asset_factory
):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31, "sep38"])
    offchain_asset = OffChainAsset.objects.create(scheme="test", identifier="test")
    ExchangePair.objects.create(
        sell_asset=asset.asset_identification_format,
        buy_asset=offchain_asset.asset_identification_format,
    )
    quote = Quote.objects.create(
        id=str(uuid.uuid4()),
        stellar_account="test source address",
        type=Quote.TYPE.indicative,
        sell_asset=asset.asset_identification_format,
        buy_asset=offchain_asset.asset_identification_format,
        sell_amount=100,
        buy_amount=100,
        price=1,
        expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
    )
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "quote_id": quote.id,
            "destination_asset": offchain_asset.asset_identification_format,
            "amount": 100,
            "sender_id": "123",
            "receiver_id": "456",
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 400, response.content
    assert body == {"error": "quote not found"}
    mock_sep31_ri.process_post_request.assert_not_called()


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration")
def test_send_firm_quote_needs_matching_id(mock_sep31_ri, client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31, "sep38"])
    offchain_asset = OffChainAsset.objects.create(scheme="test", identifier="test")
    ExchangePair.objects.create(
        sell_asset=asset.asset_identification_format,
        buy_asset=offchain_asset.asset_identification_format,
    )
    Quote.objects.create(
        id=str(uuid.uuid4()),
        stellar_account="test source address",
        type=Quote.TYPE.firm,
        sell_asset=asset.asset_identification_format,
        buy_asset=offchain_asset.asset_identification_format,
        sell_amount=100,
        buy_amount=100,
        price=1,
        expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
    )
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "quote_id": str(uuid.uuid4()),
            "destination_asset": offchain_asset.asset_identification_format,
            "amount": 100,
            "sender_id": "123",
            "receiver_id": "456",
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 400, response.content
    assert body == {"error": "quote not found"}
    mock_sep31_ri.process_post_request.assert_not_called()


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration")
def test_send_firm_quote_needs_matching_buy_asset(
    mock_sep31_ri, client, usd_asset_factory
):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31, "sep38"])
    offchain_asset = OffChainAsset.objects.create(scheme="test", identifier="test")
    ExchangePair.objects.create(
        sell_asset=asset.asset_identification_format,
        buy_asset=offchain_asset.asset_identification_format,
    )
    quote = Quote.objects.create(
        id=str(uuid.uuid4()),
        stellar_account="test source address",
        type=Quote.TYPE.firm,
        sell_asset=asset.asset_identification_format,
        buy_asset="not:test",
        sell_amount=100,
        buy_amount=100,
        price=1,
        expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
    )
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "quote_id": str(quote.id),
            "destination_asset": offchain_asset.asset_identification_format,
            "amount": 100,
            "sender_id": "123",
            "receiver_id": "456",
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 400, response.content
    assert body == {
        "error": "quote 'buy_asset' does not match 'destination_asset' parameter"
    }
    mock_sep31_ri.process_post_request.assert_not_called()


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration")
def test_send_firm_quote_needs_matching_auth(mock_sep31_ri, client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31, "sep38"])
    offchain_asset = OffChainAsset.objects.create(scheme="test", identifier="test")
    ExchangePair.objects.create(
        sell_asset=asset.asset_identification_format,
        buy_asset=offchain_asset.asset_identification_format,
    )
    quote = Quote.objects.create(
        id=str(uuid.uuid4()),
        stellar_account="not source address",
        type=Quote.TYPE.firm,
        sell_asset=asset.asset_identification_format,
        buy_asset=offchain_asset.asset_identification_format,
        sell_amount=100,
        buy_amount=100,
        price=1,
        expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
    )
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "quote_id": str(quote.id),
            "destination_asset": offchain_asset.asset_identification_format,
            "amount": 100,
            "sender_id": "123",
            "receiver_id": "456",
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 400, response.content
    assert body == {"error": "quote not found"}
    mock_sep31_ri.process_post_request.assert_not_called()


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration")
def test_send_firm_quote_expired(mock_sep31_ri, client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31, "sep38"])
    offchain_asset = OffChainAsset.objects.create(scheme="test", identifier="test")
    ExchangePair.objects.create(
        sell_asset=asset.asset_identification_format,
        buy_asset=offchain_asset.asset_identification_format,
    )
    quote = Quote.objects.create(
        id=str(uuid.uuid4()),
        stellar_account="test source address",
        type=Quote.TYPE.firm,
        sell_asset=asset.asset_identification_format,
        buy_asset=offchain_asset.asset_identification_format,
        sell_amount=100,
        buy_amount=100,
        price=1,
        expires_at=datetime.now(timezone.utc) - timedelta(hours=1),
    )
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "quote_id": str(quote.id),
            "destination_asset": offchain_asset.asset_identification_format,
            "amount": 100,
            "sender_id": "123",
            "receiver_id": "456",
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 400, response.content
    assert body == {"error": "quote has expired"}
    mock_sep31_ri.process_post_request.assert_not_called()


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration")
def test_send_firm_quote_sell_asset_no_match(mock_sep31_ri, client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31, "sep38"])
    offchain_asset = OffChainAsset.objects.create(scheme="test", identifier="test")
    ExchangePair.objects.create(
        sell_asset=asset.asset_identification_format,
        buy_asset=offchain_asset.asset_identification_format,
    )
    quote = Quote.objects.create(
        id=str(uuid.uuid4()),
        stellar_account="test source address",
        type=Quote.TYPE.firm,
        sell_asset="stellar:not:match",
        buy_asset=offchain_asset.asset_identification_format,
        sell_amount=100,
        buy_amount=100,
        price=1,
        expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
    )
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "quote_id": str(quote.id),
            "destination_asset": offchain_asset.asset_identification_format,
            "amount": 100,
            "sender_id": "123",
            "receiver_id": "456",
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 400, response.content
    assert body == {
        "error": "quote 'sell_asset' does not match 'asset_code' and 'asset_issuer' parameters"
    }
    mock_sep31_ri.process_post_request.assert_not_called()


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration")
def test_send_firm_quote_buy_asset_no_match(mock_sep31_ri, client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31, "sep38"])
    offchain_asset = OffChainAsset.objects.create(scheme="test", identifier="test")
    ExchangePair.objects.create(
        sell_asset=asset.asset_identification_format,
        buy_asset=offchain_asset.asset_identification_format,
    )
    quote = Quote.objects.create(
        id=str(uuid.uuid4()),
        stellar_account="test source address",
        type=Quote.TYPE.firm,
        sell_asset=asset.asset_identification_format,
        buy_asset="not:test",
        sell_amount=100,
        buy_amount=100,
        price=1,
        expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
    )
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "quote_id": str(quote.id),
            "destination_asset": offchain_asset.asset_identification_format,
            "amount": 100,
            "sender_id": "123",
            "receiver_id": "456",
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 400, response.content
    assert body == {
        "error": "quote 'buy_asset' does not match 'destination_asset' parameter"
    }
    mock_sep31_ri.process_post_request.assert_not_called()


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration")
def test_send_firm_quote_amount_no_match(mock_sep31_ri, client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31, "sep38"])
    offchain_asset = OffChainAsset.objects.create(scheme="test", identifier="test")
    ExchangePair.objects.create(
        sell_asset=asset.asset_identification_format,
        buy_asset=offchain_asset.asset_identification_format,
    )
    quote = Quote.objects.create(
        id=str(uuid.uuid4()),
        stellar_account="test source address",
        type=Quote.TYPE.firm,
        sell_asset=asset.asset_identification_format,
        buy_asset=offchain_asset.asset_identification_format,
        sell_amount=100,
        buy_amount=100,
        price=1,
        expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
    )
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "quote_id": str(quote.id),
            "destination_asset": offchain_asset.asset_identification_format,
            "amount": 1000,
            "sender_id": "123",
            "receiver_id": "456",
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 400, response.content
    assert body == {"error": "quote amount does not patch 'amount' parameter"}
    mock_sep31_ri.process_post_request.assert_not_called()


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration")
def test_send_firm_quote_no_offchain_asset(mock_sep31_ri, client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31, "sep38"])
    ExchangePair.objects.create(
        sell_asset=asset.asset_identification_format, buy_asset="test:test"
    )
    quote = Quote.objects.create(
        id=str(uuid.uuid4()),
        stellar_account="test source address",
        type=Quote.TYPE.firm,
        sell_asset=asset.asset_identification_format,
        buy_asset="test:test",
        sell_amount=100,
        buy_amount=100,
        price=1,
        expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
    )
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "quote_id": str(quote.id),
            "destination_asset": "test:test",
            "amount": 100,
            "sender_id": "123",
            "receiver_id": "456",
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 400, response.content
    assert body == {"error": "invalid 'destination_asset'"}
    mock_sep31_ri.process_post_request.assert_not_called()


@pytest.mark.django_db
def test_auth_check(client):
    response = client.post(transaction_endpoint, {})
    assert response.status_code == 403
    assert "error" in response.json()


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch(
    "polaris.sep31.transactions.registered_sep31_receiver_integration",
    success_send_integration,
)
def test_missing_category(client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31])
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "amount": 100,
            "fields": {},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 400
    assert body["error"] == "transaction_info_needed"
    assert body["fields"] == {
        "transaction": success_send_integration.info()["fields"]["transaction"]
    }


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch(
    "polaris.sep31.transactions.registered_sep31_receiver_integration",
    success_send_integration,
)
def test_missing_field(client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31])
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "amount": 100,
            "fields": {"transaction": {}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 400
    assert body["error"] == "transaction_info_needed"
    assert body["fields"] == {
        "transaction": {
            "bank_account": success_send_integration.info()["fields"]["transaction"][
                "bank_account"
            ]
        }
    }


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch(
    "polaris.sep31.transactions.registered_sep31_receiver_integration",
    success_send_integration,
)
def test_extra_field(client, usd_asset_factory):
    """
    Don't return 400 on extra fields passed
    """
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31])
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "amount": 100,
            "fields": {
                "transaction": {"bank_account": "fake account", "extra": "field"},
            },
        },
        content_type="application/json",
    )
    assert response.status_code == 201


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch(
    "polaris.sep31.transactions.registered_sep31_receiver_integration",
    success_send_integration,
)
def test_extra_category(client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31])
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "amount": 100,
            "fields": {
                "extra": {"category": "value"},
                "transaction": {"bank_account": "fake account"},
            },
        },
        content_type="application/json",
    )
    assert response.status_code == 400


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch(
    "polaris.sep31.transactions.registered_sep31_receiver_integration",
    success_send_integration,
)
def test_missing_amount(client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31])
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 400
    assert "error" in body
    assert "amount" in body["error"]


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch(
    "polaris.sep31.transactions.registered_sep31_receiver_integration",
    success_send_integration,
)
def test_missing_asset_code(client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31])
    response = client.post(
        transaction_endpoint,
        {
            "asset_issuer": asset.issuer,
            "amount": 100,
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 400
    assert "error" in body
    assert "asset_code" in body["error"]


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch(
    "polaris.sep31.transactions.registered_sep31_receiver_integration",
    success_send_integration,
)
def test_bad_asset_issuer(client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31])
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": Keypair.random().public_key,
            "amount": 100,
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 400
    assert "error" in body
    assert "asset_issuer" in body["error"]


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch(
    "polaris.sep31.transactions.registered_sep31_receiver_integration",
    success_send_integration,
)
def test_bad_asset_code(client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31])
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": "FAKE",
            "asset_issuer": asset.issuer,
            "amount": 100,
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 400
    assert "error" in body
    assert "asset_code" in body["error"]


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch(
    "polaris.sep31.transactions.registered_sep31_receiver_integration",
    success_send_integration,
)
def test_large_amount(client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31])
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "amount": 10000,
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 400
    assert "error" in body
    assert "amount" in body["error"]


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch(
    "polaris.sep31.transactions.registered_sep31_receiver_integration",
    success_send_integration,
)
def test_small_amount(client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31])
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "amount": 0.001,
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 400
    assert "error" in body
    assert "amount" in body["error"]


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch(
    "polaris.sep31.transactions.registered_sep31_receiver_integration",
    success_send_integration,
)
def test_transaction_created(client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31])
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "amount": 100,
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    t = Transaction.objects.filter(id=body["id"]).first()
    assert response.status_code == 201
    assert t
    assert t.amount_in == 100
    assert t.stellar_account == "test source address"
    assert t.memo
    assert t.memo_type == Transaction.MEMO_TYPES.hash
    assert t.kind == Transaction.KIND.send
    assert t.status == Transaction.STATUS.pending_sender
    assert t.protocol == Transaction.PROTOCOL.sep31
    assert t.receiving_anchor_account == asset.distribution_account
    assert t.asset == asset


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch(
    "polaris.sep31.transactions.registered_sep31_receiver_integration",
    success_send_integration,
)
def test_unsupported_lang(client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31])
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "lang": "es",
            "amount": 100,
            "fields": {
                "receiver": {"first_name": "first", "last_name": "last"},
                "sender": {"first_name": "first", "last_name": "last"},
                "transaction": {"bank_account": "fake account"},
            },
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 400
    assert "error" in body
    assert "lang" in body["error"]


@pytest.mark.django_db
@patch(
    "polaris.sep31.transactions.registered_sep31_receiver_integration",
    success_send_integration,
)
def test_bad_customer_info_needed():
    with pytest.raises(ValueError) as e:
        process_post_response(
            {
                "error": "customer_info_needed",
                "fields": {"not in expected format": True},
            },
            Mock(id=uuid.uuid4()),
        )
    assert "fields" in str(e.value)


@pytest.mark.django_db
@patch(
    "polaris.sep31.transactions.registered_sep31_receiver_integration",
    success_send_integration,
)
def test_bad_value_for_category_in_response():
    with pytest.raises(ValueError):
        process_post_response(
            {"error": "customer_info_needed", "fields": {"sender": True}},
            Mock(id=uuid.uuid4()),
        )


@pytest.mark.django_db
@patch(
    "polaris.sep31.transactions.registered_sep31_receiver_integration",
    success_send_integration,
)
def test_bad_field_in_response():
    with pytest.raises(ValueError):
        process_post_response(
            {
                "error": "customer_info_needed",
                "fields": {
                    "sender": {"not_first_name": {"description": "a description"}}
                },
            },
            Mock(id=uuid.uuid4()),
        )


@pytest.mark.django_db
@patch(
    "polaris.sep31.transactions.registered_sep31_receiver_integration",
    success_send_integration,
)
def test_bad_field_value_in_response():
    with pytest.raises(ValueError):
        process_post_response(
            {
                "error": "customer_info_needed",
                "fields": {
                    "sender": {
                        "first_name": {
                            # missing description
                            "example": "an example"
                        }
                    }
                },
            },
            Mock(id=uuid.uuid4()),
        )


def bad_process_send_request(token, request, params, transaction):
    transaction.save()
    return {
        "error": "customer_info_needed",
        "fields": {"transaction": {"bank_account": {"description": "bank account"}}},
    }


bad_save_integration = Mock(
    info=Mock(
        return_value={
            "fields": {
                "transaction": {"bank_account": {"description": "bank account"}}
            },
        }
    ),
    process_post_request=bad_process_send_request,
)


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch(
    "polaris.sep31.transactions.registered_sep31_receiver_integration",
    bad_save_integration,
)
def test_bad_save(client, usd_asset_factory):
    asset = usd_asset_factory(protocols=[Transaction.PROTOCOL.sep31])
    response = client.post(
        transaction_endpoint,
        {
            "asset_code": asset.code,
            "asset_issuer": asset.issuer,
            "amount": 100,
            "fields": {"transaction": {"bank_account": "fake account"}},
        },
        content_type="application/json",
    )
    body = response.json()
    assert response.status_code == 500
    assert "error" in body



import pytest
import uuid
from unittest.mock import patch, Mock
from datetime import datetime

from polaris.tests.helpers import mock_check_auth_success
from polaris.models import Transaction, Quote
from polaris.sep31.serializers import SEP31TransactionSerializer


endpoint = "/sep31/transactions/"


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration", Mock())
def test_successful_call(client, acc1_usd_deposit_transaction_factory):
    transaction = acc1_usd_deposit_transaction_factory(
        protocol=Transaction.PROTOCOL.sep31,
    )
    # The above transaction was created in memory, not retrieved from the DB.
    # The Decimal fields are altered to allow 7 decimals of precision once they
    # are inserted, so the serialization of the above transaction as it is won't match
    # the response json unless its Decimals fields are refreshed.
    transaction.refresh_from_db()
    # stellar account has to match auth token
    transaction.stellar_account = "test source address"
    transaction.quote = Quote.objects.create(
        id=str(uuid.uuid4()),
        stellar_account=transaction.stellar_account,
        type=Quote.TYPE.indicative,
        sell_asset=transaction.asset.asset_identification_format,
        buy_asset="iso4217:USD",
        sell_amount=transaction.amount_in,
    )
    transaction.fee_asset = transaction.asset.asset_identification_format
    transaction.save()
    response = client.get(endpoint + str(transaction.id))
    serialization = {"transaction": SEP31TransactionSerializer(transaction).data}
    assert response.status_code == 200
    assert (
        response.json()
        == serialization
        == {
            "transaction": {
                "id": str(transaction.id),
                "status": "pending_sender",
                "status_eta": 3600,
                "amount_in_asset": transaction.quote.sell_asset,
                "amount_in": "18.34",
                "amount_out_asset": transaction.quote.buy_asset,
                "amount_out": "18.24",
                "amount_fee_asset": transaction.fee_asset,
                "amount_fee": "0.10",
                "started_at": datetime.isoformat(transaction.started_at).replace(
                    "+00:00", "Z"
                ),
                "completed_at": None,
                "stellar_transaction_id": None,
                "external_transaction_id": "2dd16cb409513026fbe7defc0c6f826c2d2c65c3da993f747d09bf7dafd31093",
                "refunded": False,
                "stellar_account_id": None,
                "stellar_memo": None,
                "stellar_memo_type": "text",
                "required_info_updates": None,
                "required_info_message": None,
            }
        }
    )


def test_no_auth(client):
    # No need for id arg
    response = client.get(endpoint + "test")
    assert response.status_code == 403


mock_valid_sending_anchor = Mock(valid_sending_anchor=Mock(return_value=False))


@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch(
    "polaris.sep31.transactions.registered_sep31_receiver_integration",
    mock_valid_sending_anchor,
)
def test_invalid_anchor(client):
    response = client.get(endpoint + "test")
    assert response.status_code == 403
    mock_valid_sending_anchor.valid_sending_anchor.assert_called()


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration", Mock())
def test_no_transaction(client):
    response = client.get(endpoint + str(uuid.uuid4()))
    assert response.status_code == 404


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration", Mock())
def test_bad_id(client):
    response = client.get(endpoint + "notauuid")
    assert response.status_code == 404


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
@patch("polaris.sep31.transactions.registered_sep31_receiver_integration", Mock())
def test_no_id(client):
    response = client.get(endpoint)
    assert response.status_code == 404



import uuid

import pytest
import json
from datetime import datetime, timezone, timedelta
from uuid import UUID
from unittest.mock import patch
from decimal import Decimal

from stellar_sdk import Keypair

from polaris.sep38.quote import validate_quote_provided
from polaris.tests.helpers import mock_check_auth_success
from polaris.models import DeliveryMethod, Asset, OffChainAsset, ExchangePair, Quote
from polaris.settings import DATETIME_FORMAT


ENDPOINT = "/sep38/quote"
code_path = "polaris.sep38.quote"


def default_data():
    usd_stellar = Asset.objects.create(
        code="usd", issuer=Keypair.random().public_key, sep38_enabled=True
    )
    brl_offchain = OffChainAsset.objects.create(
        scheme="iso4217", identifier="BRL", country_codes="BRA"
    )
    delivery_methods = [
        DeliveryMethod.objects.create(
            type=DeliveryMethod.TYPE.buy,
            name="cash_pickup",
            description="cash pick-up",
        ),
        DeliveryMethod.objects.create(
            type=DeliveryMethod.TYPE.sell,
            name="cash_dropoff",
            description="cash drop-off",
        ),
    ]
    brl_offchain.delivery_methods.add(*delivery_methods)
    pair = ExchangePair.objects.create(
        buy_asset=brl_offchain.asset_identification_format,
        sell_asset=usd_stellar.asset_identification_format,
    )
    return {
        "stellar_assets": [usd_stellar],
        "offchain_assets": [brl_offchain],
        "exchange_pairs": [pair],
        "delivery_methods": delivery_methods,
    }


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_success_no_optional_params(mock_rqi, client):
    data = default_data()

    def mock_post_quote(token, request, quote, *args, **kwargs) -> Quote:
        quote.price = Decimal("2.12")
        quote.expires_at = quote.requested_expire_after or datetime.now(
            timezone.utc
        ) + timedelta(hours=1)
        if quote.sell_asset.startswith("stellar"):
            quote.buy_delivery_method = data["delivery_methods"][0]
        else:
            quote.sell_delivery_method = data["delivery_methods"][1]
        return quote

    mock_rqi.post_quote = mock_post_quote

    response = client.post(
        ENDPOINT,
        json.dumps(
            {
                "sell_asset": data["stellar_assets"][0].asset_identification_format,
                "buy_asset": data["offchain_assets"][0].asset_identification_format,
                "buy_amount": 100,
                "buy_delivery_method": "cash_pickup",
            }
        ),
        content_type="application/json",
    )
    assert response.status_code == 201, response.content
    body = response.json()
    quote_id = UUID(body.pop("id"))
    datetime.strptime(body.pop("expires_at"), DATETIME_FORMAT)
    assert body == {
        "price": "2.12",
        "buy_amount": "100.00",
        "sell_amount": "212.00",
        "sell_asset": data["stellar_assets"][0].asset_identification_format,
        "buy_asset": data["offchain_assets"][0].asset_identification_format,
    }
    q = Quote.objects.get(id=quote_id)
    assert q
    assert q.price == Decimal("2.12")
    assert q.sell_amount == Decimal("212")


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_success_country_code_buy_delivery_method_expire_after(
    mock_rqi, client
):
    data = default_data()

    def mock_post_quote(token, request, quote, *args, **kwargs) -> Quote:
        quote.price = Decimal("2.12")
        quote.expires_at = quote.requested_expire_after or datetime.now(
            timezone.utc
        ) + timedelta(hours=1)
        if not quote.buy_delivery_method and quote.sell_asset.startswith("stellar"):
            quote.buy_delivery_method = data["delivery_methods"][0]
        elif not quote.sell_delivery_method and quote.buy_asset.startswith("stellar"):
            quote.sell_delivery_method = data["delivery_methods"][1]
        return quote

    mock_rqi.post_quote = mock_post_quote

    expire_after = datetime.now(timezone.utc) + timedelta(hours=24)
    response = client.post(
        ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "sell_amount": 100,
            "buy_asset": data["offchain_assets"][0].asset_identification_format,
            "country_code": "BRA",
            "buy_delivery_method": "cash_pickup",
            "expire_after": expire_after.strftime(DATETIME_FORMAT),
        },
        content_type="application/json",
    )
    assert response.status_code == 201, response.content
    body = response.json()
    quote_id = UUID(body.pop("id"))
    datetime.strptime(body.pop("expires_at"), DATETIME_FORMAT)
    assert body == {
        "price": "2.12",
        "sell_amount": "100.00",
        "buy_amount": str(round(Decimal(100) / Decimal("2.12"), 2)),
        "sell_asset": data["stellar_assets"][0].asset_identification_format,
        "buy_asset": data["offchain_assets"][0].asset_identification_format,
    }
    assert Quote.objects.get(id=quote_id)


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_success_country_code_sell_delivery_method(mock_rqi, client):
    data = default_data()
    # swap exchange pair
    pair = data["exchange_pairs"][0]
    pair.sell_asset, pair.buy_asset = pair.buy_asset, pair.sell_asset
    pair.save()

    def mock_post_quote(token, request, quote, *args, **kwargs) -> Quote:
        quote.price = Decimal("2.12")
        quote.expires_at = quote.requested_expire_after or datetime.now(
            timezone.utc
        ) + timedelta(hours=1)
        if not quote.buy_delivery_method and quote.sell_asset.startswith("stellar"):
            quote.buy_delivery_method = data["delivery_methods"][0]
        elif not quote.sell_delivery_method and quote.buy_asset.startswith("stellar"):
            quote.sell_delivery_method = data["delivery_methods"][1]
        return quote

    mock_rqi.post_quote = mock_post_quote

    response = client.post(
        ENDPOINT,
        {
            "sell_asset": data["offchain_assets"][0].asset_identification_format,
            "buy_asset": data["stellar_assets"][0].asset_identification_format,
            "buy_amount": 100,
            "country_code": "BRA",
            "sell_delivery_method": "cash_dropoff",
        },
        content_type="application/json",
    )
    assert response.status_code == 201, response.content
    body = response.json()
    quote_id = UUID(body.pop("id"))
    datetime.strptime(body.pop("expires_at"), DATETIME_FORMAT)
    assert body == {
        "price": "2.12",
        "buy_amount": "100.00",
        "sell_amount": "212.00",
        "buy_asset": data["stellar_assets"][0].asset_identification_format,
        "sell_asset": data["offchain_assets"][0].asset_identification_format,
    }
    assert Quote.objects.get(id=quote_id)


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_success_no_optional_params_swap_exchange_pair(mock_rqi, client):
    data = default_data()
    pair = data["exchange_pairs"][0]
    pair.sell_asset, pair.buy_asset = pair.buy_asset, pair.sell_asset
    pair.save()

    def mock_post_quote(token, request, quote, *args, **kwargs) -> Quote:
        quote.price = Decimal("2.12")
        quote.expires_at = quote.requested_expire_after or datetime.now(
            timezone.utc
        ) + timedelta(hours=1)
        if not quote.buy_delivery_method and quote.sell_asset.startswith("stellar"):
            quote.buy_delivery_method = data["delivery_methods"][0]
        elif not quote.sell_delivery_method and quote.buy_asset.startswith("stellar"):
            quote.sell_delivery_method = data["delivery_methods"][1]
        return quote

    mock_rqi.post_quote = mock_post_quote

    response = client.post(
        ENDPOINT,
        {
            "sell_asset": data["offchain_assets"][0].asset_identification_format,
            "sell_amount": 100,
            "buy_asset": data["stellar_assets"][0].asset_identification_format,
            "sell_delivery_method": "cash_dropoff",
        },
        content_type="application/json",
    )
    assert response.status_code == 201, response.content
    body = response.json()
    quote_id = UUID(body.pop("id"))
    datetime.strptime(body.pop("expires_at"), DATETIME_FORMAT)
    assert body == {
        "price": "2.12",
        "sell_amount": "100.00",
        "buy_amount": str(round(Decimal(100) / Decimal("2.12"), 2)),
        "buy_asset": data["stellar_assets"][0].asset_identification_format,
        "sell_asset": data["offchain_assets"][0].asset_identification_format,
    }
    assert Quote.objects.get(id=quote_id)


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_failure_both_amounts(mock_rqi, client):
    data = default_data()

    response = client.post(
        ENDPOINT,
        json.dumps(
            {
                "sell_asset": data["stellar_assets"][0].asset_identification_format,
                "buy_asset": data["offchain_assets"][0].asset_identification_format,
                "buy_amount": 100,
                "sell_amount": 100,
                "buy_delivery_method": "cash_pickup",
            }
        ),
        content_type="application/json",
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "'sell_amount' or 'buy_amount' is required, but both is invalid"
    }
    mock_rqi.post_quote.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_failure_no_exchange_pairs(mock_rqi, client):
    data = default_data()
    response = client.post(
        ENDPOINT,
        {
            "sell_asset": data["offchain_assets"][0].asset_identification_format,
            "buy_asset": data["stellar_assets"][0].asset_identification_format,
            "buy_amount": 100,
            "sell_delivery_method": "cash_dropoff",
        },
        content_type="application/json",
    )
    assert response.status_code == 400, response.content
    assert response.json() == {"error": "unsupported asset pair"}
    mock_rqi.post_quote.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_failure_missing_sell_asset(mock_rqi, client):
    data = default_data()
    response = client.post(
        ENDPOINT,
        {
            "sell_amount": 100,
            "buy_asset": data["offchain_assets"][0].asset_identification_format,
            "buy_delivery_method": "cash_pickup",
        },
        content_type="application/json",
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "missing required parameters. Required: sell_asset, buy_asset"
    }
    mock_rqi.post_quote.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_failure_missing_buy_asset(mock_rqi, client):
    data = default_data()
    response = client.post(
        ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "buy_amount": 100,
            "buy_delivery_method": "cash_pickup",
        },
        content_type="application/json",
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "missing required parameters. Required: sell_asset, buy_asset"
    }
    mock_rqi.post_quote.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_failure_both_delivery_methods(mock_rqi, client):
    data = default_data()
    response = client.post(
        ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "sell_amount": 100,
            "buy_asset": data["offchain_assets"][0].asset_identification_format,
            "buy_delivery_method": "cash_pickup",
            "sell_delivery_method": "cash_dropoff",
        },
        content_type="application/json",
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "'buy_delivery_method' or 'sell_delivery_method' is required, but both is invalid"
    }
    mock_rqi.post_quote.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_failure_sell_stellar_with_sell_delivery_method(mock_rqi, client):
    data = default_data()
    response = client.post(
        ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "sell_amount": 100,
            "buy_asset": data["offchain_assets"][0].asset_identification_format,
            "sell_delivery_method": "cash_dropoff",
        },
        content_type="application/json",
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "unexpected 'sell_delivery_method', client intends to sell a Stellar asset"
    }
    mock_rqi.post_quote.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_failure_sell_offchain_with_buy_delivery_method(mock_rqi, client):
    data = default_data()

    # swap exchange pair
    pair = data["exchange_pairs"][0]
    pair.sell_asset, pair.buy_asset = pair.buy_asset, pair.sell_asset
    pair.save()

    response = client.post(
        ENDPOINT,
        {
            "sell_asset": data["offchain_assets"][0].asset_identification_format,
            "buy_asset": data["stellar_assets"][0].asset_identification_format,
            "buy_amount": 100,
            "buy_delivery_method": "cash_pickup",
        },
        content_type="application/json",
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "unexpected 'buy_delivery_method', client intends to buy a Stellar asset"
    }
    mock_rqi.post_quote.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_failure_bad_sell_stellar_format(mock_rqi, client):
    data = default_data()
    response = client.post(
        ENDPOINT,
        {
            "sell_asset": f"stellar:USDC",
            "buy_asset": data["offchain_assets"][0].asset_identification_format,
            "buy_amount": 100,
            "buy_delivery_method": "cash_pickup",
        },
        content_type="application/json",
    )
    assert response.status_code == 400, response.content
    assert response.json() == {"error": "invalid 'sell_asset' format"}
    mock_rqi.post_quote.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_failure_bad_buy_stellar_format(mock_rqi, client):
    data = default_data()
    response = client.post(
        ENDPOINT,
        {
            "buy_asset": f"stellar:USDC",
            "sell_asset": data["offchain_assets"][0].asset_identification_format,
            "sell_amount": 100,
            "sell_delivery_method": "cash_dropoff",
        },
        content_type="application/json",
    )
    assert response.status_code == 400, response.content
    assert response.json() == {"error": "invalid 'buy_asset' format"}
    mock_rqi.post_quote.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_failure_bad_sell_offchain_format(mock_rqi, client):
    data = default_data()
    response = client.post(
        ENDPOINT,
        {
            "sell_asset": f"USD",
            "sell_amount": 100,
            "buy_asset": data["stellar_assets"][0].asset_identification_format,
            "sell_delivery_method": "cash_dropoff",
        },
        content_type="application/json",
    )
    assert response.status_code == 400, response.content
    assert response.json() == {"error": "invalid 'sell_asset' format"}
    mock_rqi.post_quote.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_failure_bad_buy_offchain_format(mock_rqi, client):
    data = default_data()
    response = client.post(
        ENDPOINT,
        {
            "buy_asset": f"USD",
            "buy_amount": 100,
            "sell_asset": data["offchain_assets"][0].asset_identification_format,
            "sell_delivery_method": "cash_dropoff",
        },
        content_type="application/json",
    )
    assert response.status_code == 400, response.content
    assert response.json() == {"error": "invalid 'buy_asset' format"}
    mock_rqi.post_quote.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_failure_stellar_asset_not_found(mock_rqi, client):
    data = default_data()

    # delete stellar asset from DB
    data["stellar_assets"][0].delete()

    response = client.post(
        ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "buy_asset": data["offchain_assets"][0].asset_identification_format,
            "buy_amount": 100,
            "buy_delivery_method": "cash_dropoff",
        },
        content_type="application/json",
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "no 'sell_asset' for 'delivery_method' and 'country_code' specificed"
    }
    mock_rqi.post_quote.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_failure_offchain_asset_not_found(mock_rqi, client):
    data = default_data()

    # swap exchange pair
    pair = data["exchange_pairs"][0]
    pair.sell_asset, pair.buy_asset = pair.buy_asset, pair.sell_asset
    pair.save()
    # delete offchain asset from DB
    data["offchain_assets"][0].delete()

    response = client.post(
        ENDPOINT,
        {
            "sell_asset": data["offchain_assets"][0].asset_identification_format,
            "buy_asset": data["stellar_assets"][0].asset_identification_format,
            "buy_amount": 100,
            "sell_delivery_method": "cash_dropoff",
        },
        content_type="application/json",
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "no 'sell_asset' for 'delivery_method' and 'country_code' specificed"
    }
    mock_rqi.post_quote.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_failure_bad_buy_delivery_method(mock_rqi, client):
    data = default_data()

    response = client.post(
        ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "buy_asset": data["offchain_assets"][0].asset_identification_format,
            "buy_amount": 100,
            "buy_delivery_method": "bad_delivery_method",
        },
        content_type="application/json",
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "unable to find 'buy_asset' using the following filters: 'country_code', 'buy_delivery_method'"
    }
    mock_rqi.post_quote.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_failure_bad_country_code(mock_rqi, client):
    data = default_data()

    response = client.post(
        ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "sell_amount": 100,
            "buy_asset": data["offchain_assets"][0].asset_identification_format,
            "country_code": "TEST",
            "buy_delivery_method": "cash_pickup",
        },
        content_type="application/json",
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "unable to find 'buy_asset' using the following filters: 'country_code', 'buy_delivery_method'"
    }
    mock_rqi.post_quote.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_failure_bad_sell_delivery_method(mock_rqi, client):
    data = default_data()
    # swap exchange pair
    pair = data["exchange_pairs"][0]
    pair.sell_asset, pair.buy_asset = pair.buy_asset, pair.sell_asset
    pair.save()

    response = client.post(
        ENDPOINT,
        {
            "sell_asset": data["offchain_assets"][0].asset_identification_format,
            "buy_asset": data["stellar_assets"][0].asset_identification_format,
            "buy_amount": 100,
            "sell_delivery_method": "bad_delivery_method",
        },
        content_type="application/json",
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "no 'sell_asset' for 'delivery_method' and 'country_code' specificed"
    }
    mock_rqi.post_quote.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_failure_bad_sell_amount(mock_rqi, client):
    data = default_data()

    response = client.post(
        ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "sell_amount": "test",
            "buy_asset": data["offchain_assets"][0].asset_identification_format,
            "buy_delivery_method": "cash_pickup",
        },
        content_type="application/json",
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "invalid 'buy_amount' or 'sell_amount'; Expected decimal strings."
    }
    mock_rqi.post_quote.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_failure_bad_buy_amount(mock_rqi, client):
    data = default_data()

    response = client.post(
        ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "buy_asset": data["offchain_assets"][0].asset_identification_format,
            "buy_amount": "test",
            "buy_delivery_method": "cash_pickup",
        },
        content_type="application/json",
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "invalid 'buy_amount' or 'sell_amount'; Expected decimal strings."
    }
    mock_rqi.post_quote.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_failure_bad_expires_at(mock_rqi, client):
    data = default_data()

    response = client.post(
        ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "buy_asset": data["offchain_assets"][0].asset_identification_format,
            "buy_amount": 100,
            "expire_after": "test",
            "buy_delivery_method": "cash_pickup",
        },
        content_type="application/json",
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "invalid 'expire_after' string format. Expected UTC ISO 8601 datetime string."
    }
    mock_rqi.post_quote.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_failure_expires_at_in_past(mock_rqi, client):
    data = default_data()

    response = client.post(
        ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "sell_amount": 100,
            "buy_asset": data["offchain_assets"][0].asset_identification_format,
            "expire_after": (datetime.now(timezone.utc) - timedelta(hours=1)).strftime(
                DATETIME_FORMAT
            ),
            "buy_delivery_method": "cash_pickup",
        },
        content_type="application/json",
    )
    assert response.status_code == 400, response.content
    assert response.json() == {
        "error": "invalid 'expire_after' datetime. Expected future datetime."
    }
    mock_rqi.post_quote.assert_not_called()


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_failure_anchor_raises_value_error(mock_rqi, client):
    data = default_data()
    mock_rqi.post_quote.side_effect = ValueError("test")

    response = client.post(
        ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "sell_amount": 100,
            "buy_asset": data["offchain_assets"][0].asset_identification_format,
            "buy_delivery_method": "cash_pickup",
        },
        content_type="application/json",
    )
    assert response.status_code == 400, response.content
    assert response.json() == {"error": "test"}
    mock_rqi.post_quote.assert_called_once()
    kwargs = mock_rqi.post_quote.call_args[1]
    del kwargs["token"]
    del kwargs["request"]
    assert isinstance(kwargs["quote"], Quote)


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_failure_anchor_raises_runtime_error(mock_rqi, client):
    data = default_data()
    mock_rqi.post_quote.side_effect = RuntimeError("test")

    response = client.post(
        ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "buy_asset": data["offchain_assets"][0].asset_identification_format,
            "buy_amount": 100,
            "buy_delivery_method": "cash_pickup",
        },
        content_type="application/json",
    )
    assert response.status_code == 503, response.content
    assert response.json() == {"error": "test"}
    mock_rqi.post_quote.assert_called_once()
    kwargs = mock_rqi.post_quote.call_args[1]
    del kwargs["token"]
    del kwargs["request"]
    assert isinstance(kwargs["quote"], Quote)


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_failure_anchor_raises_unexpected_error(mock_rqi, client):
    data = default_data()
    mock_rqi.post_quote.side_effect = IndexError("test")

    with pytest.raises(Exception):
        client.post(
            ENDPOINT,
            {
                "sell_asset": data["stellar_assets"][0].asset_identification_format,
                "sell_amount": 100,
                "buy_asset": data["offchain_assets"][0].asset_identification_format,
                "buy_delivery_method": "cash_pickup",
            },
            content_type="application/json",
        )
    mock_rqi.post_quote.assert_called_once()
    kwargs = mock_rqi.post_quote.call_args[1]
    del kwargs["token"]
    del kwargs["request"]
    assert isinstance(kwargs["quote"], Quote)


@pytest.mark.django_db
@patch(f"{code_path}.rqi")
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_post_quote_failure_anchor_provides_bad_quote(mock_rqi, client):
    data = default_data()
    mock_rqi.post_quote.return_value = None

    response = client.post(
        ENDPOINT,
        {
            "sell_asset": data["stellar_assets"][0].asset_identification_format,
            "buy_asset": data["offchain_assets"][0].asset_identification_format,
            "buy_amount": 100,
            "buy_delivery_method": "cash_pickup",
        },
        content_type="application/json",
    )
    assert response.status_code == 500, response.content
    assert response.json() == {"error": "internal server error"}
    mock_rqi.post_quote.assert_called_once()
    kwargs = mock_rqi.post_quote.call_args[1]
    del kwargs["token"]
    del kwargs["request"]
    assert isinstance(kwargs["quote"], Quote)


def test_validate_quote_not_quote():
    with pytest.raises(ValueError, match="object returned is not a Quote"):
        validate_quote_provided(None, "", 2)


@pytest.mark.django_db
def test_validate_quote_sell_delivery_method():
    data = default_data()
    validate_quote_provided(
        Quote(
            type=Quote.TYPE.firm,
            buy_asset="stellar:test:test",
            sell_asset="test:test",
            buy_amount=Decimal(100),
            price=Decimal("2.12"),
            sell_delivery_method=data["offchain_assets"][0].delivery_methods.first(),
            expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
        ),
        "",
        2,
    )


@pytest.mark.django_db
def test_validate_quote_bad_price():
    data = default_data()
    with pytest.raises(
        ValueError,
        match="the price saved to Quote.price did not have the correct number of significant decimals",
    ):
        validate_quote_provided(
            Quote(
                type=Quote.TYPE.firm,
                buy_asset="stellar:test:test",
                sell_asset="test:test",
                buy_amount=Decimal(100),
                price=Decimal("2.123"),
                sell_delivery_method=data["offchain_assets"][
                    0
                ].delivery_methods.first(),
                expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
            ),
            "",
            2,
        )


@pytest.mark.django_db
def test_validate_quote_both_amounts():
    data = default_data()
    with pytest.raises(
        ValueError,
        match="polaris will calculate the amount not specified in the request",
    ):
        validate_quote_provided(
            Quote(
                type=Quote.TYPE.firm,
                buy_asset="stellar:test:test",
                sell_asset="test:test",
                buy_amount=Decimal(100),
                sell_amount=Decimal(100),
                price=Decimal("2.12"),
                sell_delivery_method=data["offchain_assets"][
                    0
                ].delivery_methods.first(),
                expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
            ),
            "",
            2,
        )


@pytest.mark.django_db
def test_validate_quote_buy_delivery_method():
    data = default_data()
    validate_quote_provided(
        Quote(
            type=Quote.TYPE.firm,
            sell_asset="stellar:test:test",
            buy_asset="test:test",
            buy_amount=Decimal(100),
            price=Decimal("2.12"),
            buy_delivery_method=data["offchain_assets"][0].delivery_methods.first(),
            expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
        ),
        "",
        2,
    )


@pytest.mark.django_db
def test_validate_quote_bad_type():
    data = default_data()
    with pytest.raises(ValueError, match="quote is not of type 'firm'"):
        validate_quote_provided(
            Quote(
                type=Quote.TYPE.indicative,
                buy_asset="stellar:test:test",
                sell_asset="test:test",
                sell_amount=Decimal(100),
                price=Decimal("2.12"),
                sell_delivery_method=data["offchain_assets"][
                    0
                ].delivery_methods.first(),
                expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
            ),
            "",
            2,
        )


@pytest.mark.django_db
def test_validate_quote_no_price():
    data = default_data()
    with pytest.raises(ValueError, match="quote must have price"):
        validate_quote_provided(
            Quote(
                type=Quote.TYPE.firm,
                buy_asset="stellar:test:test",
                sell_asset="test:test",
                buy_amount=Decimal(100),
                sell_delivery_method=data["offchain_assets"][
                    0
                ].delivery_methods.first(),
                expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
            ),
            "",
            2,
        )


@pytest.mark.django_db
def test_validate_quote_bad_amount_types():
    data = default_data()
    with pytest.raises(
        ValueError, match="quote amounts must be of type decimal.Decimal"
    ):
        validate_quote_provided(
            Quote(
                type=Quote.TYPE.firm,
                buy_asset="stellar:test:test",
                sell_asset="test:test",
                sell_amount=100,
                price=2.12,
                sell_delivery_method=data["offchain_assets"][
                    0
                ].delivery_methods.first(),
                expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
            ),
            "",
            2,
        )


@pytest.mark.django_db
def test_validate_quote_bad_amounts():
    data = default_data()
    with pytest.raises(ValueError, match="quote amounts must be positive"):
        validate_quote_provided(
            Quote(
                type=Quote.TYPE.firm,
                buy_asset="stellar:test:test",
                sell_asset="test:test",
                buy_amount=Decimal(-100),
                price=Decimal("2.12"),
                sell_delivery_method=data["offchain_assets"][
                    0
                ].delivery_methods.first(),
                expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
            ),
            "",
            2,
        )


@pytest.mark.django_db
def test_validate_quote_no_expiration():
    data = default_data()
    with pytest.raises(ValueError, match="quote must have expiration"):
        validate_quote_provided(
            Quote(
                type=Quote.TYPE.firm,
                buy_asset="stellar:test:test",
                sell_asset="test:test",
                sell_amount=Decimal(100),
                price=Decimal("2.12"),
                sell_delivery_method=data["offchain_assets"][
                    0
                ].delivery_methods.first(),
            ),
            "",
            2,
        )


@pytest.mark.django_db
def test_validate_quote_expiration_not_in_the_future():
    data = default_data()
    with pytest.raises(ValueError, match="quote expiration must be in the future"):
        validate_quote_provided(
            Quote(
                type=Quote.TYPE.firm,
                buy_asset="stellar:test:test",
                sell_asset="test:test",
                buy_amount=Decimal(100),
                price=Decimal("2.12"),
                sell_delivery_method=data["offchain_assets"][
                    0
                ].delivery_methods.first(),
                expires_at=datetime.now(timezone.utc) - timedelta(hours=1),
            ),
            "",
            2,
        )


@pytest.mark.django_db
def test_validate_quote_expiration_not_after_requested():
    data = default_data()
    with pytest.raises(
        ValueError, match="quote expiration must be at or after requested 'expire_at'"
    ):
        validate_quote_provided(
            Quote(
                type=Quote.TYPE.firm,
                buy_asset="stellar:test:test",
                sell_asset="test:test",
                sell_amount=Decimal(100),
                price=Decimal("2.12"),
                sell_delivery_method=data["offchain_assets"][
                    0
                ].delivery_methods.first(),
                expires_at=datetime.now(timezone.utc) + timedelta(minutes=20),
            ),
            (datetime.now(timezone.utc) + timedelta(hours=1)).strftime(DATETIME_FORMAT),
            2,
        )


@pytest.mark.django_db
def test_validate_quote_both_delivery_methods():
    data = default_data()
    with pytest.raises(
        ValueError,
        match="quote must have either have buy_delivery_method or sell_delivery_method'",
    ):
        validate_quote_provided(
            Quote(
                type=Quote.TYPE.firm,
                buy_asset="stellar:test:test",
                sell_asset="test:test",
                buy_amount=Decimal(100),
                price=Decimal("2.12"),
                sell_delivery_method=data["offchain_assets"][
                    0
                ].delivery_methods.first(),
                buy_delivery_method=data["offchain_assets"][0].delivery_methods.first(),
                expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
            ),
            "",
            2,
        )


def test_validate_quote_neither_delivery_methods():
    with pytest.raises(
        ValueError,
        match="quote must have either have buy_delivery_method or sell_delivery_method'",
    ):
        validate_quote_provided(
            Quote(
                type=Quote.TYPE.firm,
                buy_asset="stellar:test:test",
                sell_asset="test:test",
                sell_amount=Decimal(100),
                price=Decimal("2.12"),
                expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
            ),
            "",
            2,
        )


@pytest.mark.django_db
def test_validate_quote_bad_asset_type():
    data = default_data()
    with pytest.raises(ValueError, match="quote assets must be strings"):
        validate_quote_provided(
            Quote(
                type=Quote.TYPE.firm,
                buy_asset=1,
                sell_asset="test:test",
                buy_amount=Decimal(100),
                price=Decimal("2.12"),
                sell_delivery_method=data["offchain_assets"][
                    0
                ].delivery_methods.first(),
                expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
            ),
            "",
            2,
        )


@pytest.mark.django_db
def test_validate_quote_both_stellar_assets():
    data = default_data()
    with pytest.raises(
        ValueError, match="quote must have one stellar asset and one off chain asset"
    ):
        validate_quote_provided(
            Quote(
                type=Quote.TYPE.firm,
                buy_asset="stellar:test:test",
                sell_asset="stellar:test:test",
                buy_amount=Decimal(100),
                price=Decimal("2.12"),
                sell_delivery_method=data["offchain_assets"][
                    0
                ].delivery_methods.first(),
                expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
            ),
            "",
            2,
        )


@pytest.mark.django_db
def test_validate_quote_both_offchain_assets():
    data = default_data()
    with pytest.raises(
        ValueError, match="quote must have one stellar asset and one off chain asset"
    ):
        validate_quote_provided(
            Quote(
                type=Quote.TYPE.firm,
                buy_asset="test:test",
                sell_asset="test:test",
                sell_amount=Decimal(100),
                price=Decimal("2.12"),
                sell_delivery_method=data["offchain_assets"][
                    0
                ].delivery_methods.first(),
                expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
            ),
            "",
            2,
        )


# GET /quote/:id tests


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_quote_success(client):
    data = default_data()
    quote = Quote.objects.create(
        type=Quote.TYPE.firm,
        buy_asset=data["offchain_assets"][0].asset_identification_format,
        sell_asset=data["stellar_assets"][0].asset_identification_format,
        buy_amount=Decimal(100),
        sell_amount=Decimal(100),
        price=Decimal("2.12"),
        sell_delivery_method=data["offchain_assets"][0].delivery_methods.first(),
        expires_at=datetime.now(timezone.utc) + timedelta(hours=1),
    )
    response = client.get(f"{ENDPOINT}/{quote.id}")
    assert response.status_code == 200
    assert response.json() == {
        "id": str(quote.id),
        "expires_at": quote.expires_at.strftime(DATETIME_FORMAT),
        "buy_asset": data["offchain_assets"][0].asset_identification_format,
        "sell_asset": data["stellar_assets"][0].asset_identification_format,
        "price": "2.12",
        "sell_amount": "100.00",
        "buy_amount": "100.00",
    }


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_quote_not_found(client):
    default_data()
    response = client.get(f"{ENDPOINT}/{str(uuid.uuid4())}")
    assert response.status_code == 404, response.content
    assert response.json() == {"error": "quote not found"}


@pytest.mark.django_db
@patch("polaris.sep10.utils.check_auth", mock_check_auth_success)
def test_get_quote_not_found_bad_id(client):
    default_data()
    response = client.get(f"{ENDPOINT}/test")
    assert response.status_code == 404, response.content
    assert response.json() == {"error": "quote not found"}
